@@ -3,8 +3,8 @@ use std::borrow::Cow::{self, Borrowed, Owned};
 use std::fs;
 use std::path::{self, Path};
 
-use super::Result;
 use crate::line_buffer::LineBuffer;
+use crate::{Context, Result};
 use memchr::memchr;
 
 // TODO: let the implementers choose/find word boundaries ???
@@ -54,7 +54,15 @@ pub trait Completer {
     /// partial word to be completed.
     ///
     /// ("ls /usr/loc", 11) => Ok((3, vec!["/usr/local/"]))
-    fn complete(&self, line: &str, pos: usize) -> Result<(usize, Vec<Self::Candidate>)>;
+    fn complete(
+        &self,
+        line: &str,
+        pos: usize,
+        ctx: &Context<'_>,
+    ) -> Result<(usize, Vec<Self::Candidate>)> {
+        let _ = (line, pos, ctx);
+        Ok((0, Vec::with_capacity(0)))
+    }
     /// Updates the edited `line` with the `elected` candidate.
     fn update(&self, line: &mut LineBuffer, start: usize, elected: &str) {
         let end = line.pos();
@@ -65,10 +73,6 @@ pub trait Completer {
 impl Completer for () {
     type Candidate = String;
 
-    fn complete(&self, _line: &str, _pos: usize) -> Result<(usize, Vec<String>)> {
-        Ok((0, Vec::with_capacity(0)))
-    }
-
     fn update(&self, _line: &mut LineBuffer, _start: usize, _elected: &str) {
         unreachable!()
     }
@@ -77,8 +81,13 @@ impl Completer for () {
 impl<'c, C: ?Sized + Completer> Completer for &'c C {
     type Candidate = C::Candidate;
 
-    fn complete(&self, line: &str, pos: usize) -> Result<(usize, Vec<Self::Candidate>)> {
-        (**self).complete(line, pos)
+    fn complete(
+        &self,
+        line: &str,
+        pos: usize,
+        ctx: &Context<'_>,
+    ) -> Result<(usize, Vec<Self::Candidate>)> {
+        (**self).complete(line, pos, ctx)
     }
 
     fn update(&self, line: &mut LineBuffer, start: usize, elected: &str) {
@@ -91,8 +100,8 @@ macro_rules! box_completer {
             impl<C: ?Sized + Completer> Completer for $id<C> {
                 type Candidate = C::Candidate;
 
-                fn complete(&self, line: &str, pos: usize) -> Result<(usize, Vec<Self::Candidate>)> {
-                    (**self).complete(line, pos)
+                fn complete(&self, line: &str, pos: usize, ctx: &Context<'_>) -> Result<(usize, Vec<Self::Candidate>)> {
+                    (**self).complete(line, pos, ctx)
                 }
                 fn update(&self, line: &mut LineBuffer, start: usize, elected: &str) {
                     (**self).update(line, start, elected)
@@ -152,18 +161,8 @@ impl FilenameCompleter {
             double_quotes_special_chars: &DOUBLE_QUOTES_SPECIAL_CHARS,
         }
     }
-}
 
-impl Default for FilenameCompleter {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
-impl Completer for FilenameCompleter {
-    type Candidate = Pair;
-
-    fn complete(&self, line: &str, pos: usize) -> Result<(usize, Vec<Pair>)> {
+    pub fn complete_path(&self, line: &str, pos: usize) -> Result<(usize, Vec<Pair>)> {
         let (start, path, esc_char, break_chars, quote) =
             if let Some((idx, quote)) = find_unclosed_quote(&line[..pos]) {
                 let start = idx + 1;
@@ -194,12 +193,27 @@ impl Completer for FilenameCompleter {
     }
 }
 
+impl Default for FilenameCompleter {
+    fn default() -> Self {
+        Self::new()
+    }
+}
+
+impl Completer for FilenameCompleter {
+    type Candidate = Pair;
+
+    fn complete(&self, line: &str, pos: usize, _ctx: &Context<'_>) -> Result<(usize, Vec<Pair>)> {
+        self.complete_path(line, pos)
+    }
+}
+
 /// Remove escape char
 pub fn unescape(input: &str, esc_char: Option<char>) -> Cow<'_, str> {
-    if esc_char.is_none() {
+    let esc_char = if let Some(c) = esc_char {
+        c
+    } else {
         return Borrowed(input);
-    }
-    let esc_char = esc_char.unwrap();
+    };
     if !input.chars().any(|c| c == esc_char) {
         return Borrowed(input);
     }
@@ -242,14 +256,15 @@ pub fn escape(
     if n == 0 {
         return input; // no need to escape
     }
-    if esc_char.is_none() {
+    let esc_char = if let Some(c) = esc_char {
+        c
+    } else {
         if cfg!(windows) && quote == Quote::None {
             input.insert(0, '"'); // force double quote
             return input;
         }
         return input;
-    }
-    let esc_char = esc_char.unwrap();
+    };
     let mut result = String::with_capacity(input.len() + n);
 
     for c in input.chars() {
@@ -267,6 +282,7 @@ fn filename_complete(
     break_chars: &[u8],
     quote: Quote,
 ) -> Result<Vec<Pair>> {
+    #[cfg(feature = "with-dirs")]
     use dirs::home_dir;
     use std::env::current_dir;
 
@@ -279,12 +295,19 @@ fn filename_complete(
     let dir_path = Path::new(dir_name);
     let dir = if dir_path.starts_with("~") {
         // ~[/...]
-        if let Some(home) = home_dir() {
-            match dir_path.strip_prefix("~") {
-                Ok(rel_path) => home.join(rel_path),
-                _ => home,
+        #[cfg(feature = "with-dirs")]
+        {
+            if let Some(home) = home_dir() {
+                match dir_path.strip_prefix("~") {
+                    Ok(rel_path) => home.join(rel_path),
+                    _ => home,
+                }
+            } else {
+                dir_path.to_path_buf()
             }
-        } else {
+        }
+        #[cfg(not(feature = "with-dirs"))]
+        {
             dir_path.to_path_buf()
         }
     } else if dir_path.is_relative() {
@@ -345,8 +368,8 @@ pub fn extract_word<'l>(
     }
     let mut start = None;
     for (i, c) in line.char_indices().rev() {
-        if esc_char.is_some() && start.is_some() {
-            if esc_char.unwrap() == c {
+        if let (Some(esc_char), true) = (esc_char, start.is_some()) {
+            if esc_char == c {
                 // escaped break char
                 start = None;
                 continue;
