@@ -6,44 +6,47 @@ use std::rc::Rc;
 use unicode_segmentation::UnicodeSegmentation;
 use unicode_width::UnicodeWidthChar;
 
-use super::Result;
+use super::{Context, Helper, Result};
 use crate::highlight::Highlighter;
-use crate::hint::Hinter;
-use crate::history::{Direction, History};
+use crate::history::Direction;
 use crate::keymap::{Anchor, At, CharSearch, Cmd, Movement, RepeatCount, Word};
 use crate::keymap::{InputState, Refresher};
 use crate::line_buffer::{LineBuffer, WordAction, MAX_LINE};
-use crate::tty::{Position, RawReader, Renderer};
+use crate::tty::{Position, Renderer, Term, Terminal};
 use crate::undo::Changeset;
 
 /// Represent the state during line editing.
 /// Implement rendering.
-pub struct State<'out, 'prompt> {
-    pub out: &'out mut dyn Renderer,
+pub struct State<'out, 'prompt, H: Helper> {
+    pub out: &'out mut <Terminal as Term>::Writer,
     prompt: &'prompt str,  // Prompt to display (rl_prompt)
     prompt_size: Position, // Prompt Unicode/visible width and height
     pub line: LineBuffer,  // Edited line buffer
     pub cursor: Position,  /* Cursor position (relative to the start of the prompt
                             * for `row`) */
     pub old_rows: usize, // Number of rows used so far (from start of prompt to end of input)
-    history_index: usize, // The history index we are currently editing
     saved_line_for_history: LineBuffer, // Current edited line before history browsing
     byte_buffer: [u8; 4],
     pub changes: Rc<RefCell<Changeset>>, // changes to line, for undo/redo
-    pub hinter: Option<&'out dyn Hinter>,
-    pub highlighter: Option<&'out dyn Highlighter>,
-    no_hint: bool,        // `false` if an hint has been displayed
-    highlight_char: bool, // `true` if a char has been highlighted
+    pub helper: Option<&'out H>,
+    pub ctx: Context<'out>,   // Give access to history for `hinter`
+    pub hint: Option<String>, // last hint displayed
+    highlight_char: bool,     // `true` if a char has been highlighted
 }
 
-impl<'out, 'prompt> State<'out, 'prompt> {
+enum Info<'m> {
+    NoHint,
+    Hint,
+    Msg(Option<&'m str>),
+}
+
+impl<'out, 'prompt, H: Helper> State<'out, 'prompt, H> {
     pub fn new(
-        out: &'out mut dyn Renderer,
+        out: &'out mut <Terminal as Term>::Writer,
         prompt: &'prompt str,
-        history_index: usize,
-        hinter: Option<&'out dyn Hinter>,
-        highlighter: Option<&'out dyn Highlighter>,
-    ) -> State<'out, 'prompt> {
+        helper: Option<&'out H>,
+        ctx: Context<'out>,
+    ) -> State<'out, 'prompt, H> {
         let capacity = MAX_LINE;
         let prompt_size = out.calculate_position(prompt, Position::default());
         State {
@@ -53,21 +56,28 @@ impl<'out, 'prompt> State<'out, 'prompt> {
             line: LineBuffer::with_capacity(capacity),
             cursor: prompt_size,
             old_rows: 0,
-            history_index,
             saved_line_for_history: LineBuffer::with_capacity(capacity),
             byte_buffer: [0; 4],
             changes: Rc::new(RefCell::new(Changeset::new())),
-            hinter,
-            highlighter,
-            no_hint: true,
+            helper,
+            ctx,
+            hint: None,
             highlight_char: false,
         }
     }
 
-    pub fn next_cmd<R: RawReader>(
+    pub fn highlighter(&self) -> Option<&dyn Highlighter> {
+        if self.out.colors_enabled() {
+            self.helper.map(|h| h as &dyn Highlighter)
+        } else {
+            None
+        }
+    }
+
+    pub fn next_cmd(
         &mut self,
         input_state: &mut InputState,
-        rdr: &mut R,
+        rdr: &mut <Terminal as Term>::Reader,
         single_esc_abort: bool,
     ) -> Result<Cmd> {
         loop {
@@ -106,7 +116,7 @@ impl<'out, 'prompt> State<'out, 'prompt> {
         }
         if self.highlight_char() {
             let prompt_size = self.prompt_size;
-            self.refresh(self.prompt, prompt_size, None)?;
+            self.refresh(self.prompt, prompt_size, true, Info::NoHint)?;
         } else {
             self.out.move_cursor(self.cursor, cursor)?;
         }
@@ -114,15 +124,36 @@ impl<'out, 'prompt> State<'out, 'prompt> {
         Ok(())
     }
 
-    fn refresh(&mut self, prompt: &str, prompt_size: Position, hint: Option<String>) -> Result<()> {
+    pub fn move_cursor_at_leftmost(&mut self, rdr: &mut <Terminal as Term>::Reader) -> Result<()> {
+        self.out.move_cursor_at_leftmost(rdr)
+    }
+
+    fn refresh(
+        &mut self,
+        prompt: &str,
+        prompt_size: Position,
+        default_prompt: bool,
+        info: Info<'_>,
+    ) -> Result<()> {
+        let info = match info {
+            Info::NoHint => None,
+            Info::Hint => self.hint.as_ref().map(String::as_str),
+            Info::Msg(msg) => msg,
+        };
+        let highlighter = if self.out.colors_enabled() {
+            self.helper.map(|h| h as &dyn Highlighter)
+        } else {
+            None
+        };
         let (cursor, end_pos) = self.out.refresh_line(
             prompt,
             prompt_size,
+            default_prompt,
             &self.line,
-            hint,
+            info,
             self.cursor.row,
             self.old_rows,
-            self.highlighter,
+            highlighter,
         )?;
 
         self.cursor = cursor;
@@ -130,18 +161,17 @@ impl<'out, 'prompt> State<'out, 'prompt> {
         Ok(())
     }
 
-    fn hint(&mut self) -> Option<String> {
-        if let Some(hinter) = self.hinter {
-            let hint = hinter.hint(self.line.as_str(), self.line.pos());
-            self.no_hint = hint.is_none();
-            hint
+    pub fn hint(&mut self) {
+        if let Some(hinter) = self.helper {
+            let hint = hinter.hint(self.line.as_str(), self.line.pos(), &self.ctx);
+            self.hint = hint;
         } else {
-            None
+            self.hint = None
         }
     }
 
     fn highlight_char(&mut self) -> bool {
-        if let Some(highlighter) = self.highlighter {
+        if let Some(highlighter) = self.highlighter() {
             let highlight_char = highlighter.highlight_char(&self.line, self.line.pos());
             if highlight_char {
                 self.highlight_char = true;
@@ -159,19 +189,31 @@ impl<'out, 'prompt> State<'out, 'prompt> {
     }
 }
 
-impl<'out, 'prompt> Refresher for State<'out, 'prompt> {
+impl<'out, 'prompt, H: Helper> Refresher for State<'out, 'prompt, H> {
     fn refresh_line(&mut self) -> Result<()> {
         let prompt_size = self.prompt_size;
-        let hint = self.hint();
+        self.hint();
         self.highlight_char();
-        self.refresh(self.prompt, prompt_size, hint)
+        self.refresh(self.prompt, prompt_size, true, Info::Hint)
+    }
+
+    fn refresh_line_with_msg(&mut self, msg: Option<String>) -> Result<()> {
+        let prompt_size = self.prompt_size;
+        self.hint = None;
+        self.highlight_char();
+        self.refresh(
+            self.prompt,
+            prompt_size,
+            true,
+            Info::Msg(msg.as_ref().map(String::as_str)),
+        )
     }
 
     fn refresh_prompt_and_line(&mut self, prompt: &str) -> Result<()> {
         let prompt_size = self.out.calculate_position(prompt, Position::default());
-        let hint = self.hint();
+        self.hint();
         self.highlight_char();
-        self.refresh(prompt, prompt_size, hint)
+        self.refresh(prompt, prompt_size, false, Info::Hint)
     }
 
     fn doing_insert(&mut self) {
@@ -191,11 +233,11 @@ impl<'out, 'prompt> Refresher for State<'out, 'prompt> {
     }
 
     fn has_hint(&self) -> bool {
-        !self.no_hint
+        self.hint.is_some()
     }
 }
 
-impl<'out, 'prompt> fmt::Debug for State<'out, 'prompt> {
+impl<'out, 'prompt, H: Helper> fmt::Debug for State<'out, 'prompt, H> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("State")
             .field("prompt", &self.prompt)
@@ -204,23 +246,22 @@ impl<'out, 'prompt> fmt::Debug for State<'out, 'prompt> {
             .field("cursor", &self.cursor)
             .field("cols", &self.out.get_columns())
             .field("old_rows", &self.old_rows)
-            .field("history_index", &self.history_index)
             .field("saved_line_for_history", &self.saved_line_for_history)
             .finish()
     }
 }
 
-impl<'out, 'prompt> State<'out, 'prompt> {
+impl<'out, 'prompt, H: Helper> State<'out, 'prompt, H> {
     /// Insert the character `ch` at cursor current position.
     pub fn edit_insert(&mut self, ch: char, n: RepeatCount) -> Result<()> {
         if let Some(push) = self.line.insert(ch, n) {
             if push {
                 let prompt_size = self.prompt_size;
-                let no_previous_hint = self.no_hint;
-                let hint = self.hint();
+                let no_previous_hint = self.hint.is_none();
+                self.hint();
                 if n == 1
                     && self.cursor.col + ch.width().unwrap_or(0) < self.out.get_columns()
-                    && (hint.is_none() && no_previous_hint) // TODO refresh only current line
+                    && (self.hint.is_none() && no_previous_hint) // TODO refresh only current line
                     && !self.highlight_char()
                 {
                     // Avoid a full update of the line in the trivial case.
@@ -232,7 +273,7 @@ impl<'out, 'prompt> State<'out, 'prompt> {
                     let bits = bits.as_bytes();
                     self.out.write_and_flush(bits)
                 } else {
-                    self.refresh(self.prompt, prompt_size, hint)
+                    self.refresh(self.prompt, prompt_size, true, Info::Hint)
                 }
             } else {
                 self.refresh_line()
@@ -429,27 +470,28 @@ impl<'out, 'prompt> State<'out, 'prompt> {
 
     /// Substitute the currently edited line with the next or previous history
     /// entry.
-    pub fn edit_history_next(&mut self, history: &History, prev: bool) -> Result<()> {
+    pub fn edit_history_next(&mut self, prev: bool) -> Result<()> {
+        let history = self.ctx.history;
         if history.is_empty() {
             return Ok(());
         }
-        if self.history_index == history.len() {
+        if self.ctx.history_index == history.len() {
             if prev {
                 // Save the current edited line before overwriting it
                 self.backup();
             } else {
                 return Ok(());
             }
-        } else if self.history_index == 0 && prev {
+        } else if self.ctx.history_index == 0 && prev {
             return Ok(());
         }
         if prev {
-            self.history_index -= 1;
+            self.ctx.history_index -= 1;
         } else {
-            self.history_index += 1;
+            self.ctx.history_index += 1;
         }
-        if self.history_index < history.len() {
-            let buf = history.get(self.history_index).unwrap();
+        if self.ctx.history_index < history.len() {
+            let buf = history.get(self.ctx.history_index).unwrap();
             self.changes.borrow_mut().begin();
             self.line.update(buf, buf.len());
             self.changes.borrow_mut().end();
@@ -461,26 +503,27 @@ impl<'out, 'prompt> State<'out, 'prompt> {
     }
 
     // Non-incremental, anchored search
-    pub fn edit_history_search(&mut self, history: &History, dir: Direction) -> Result<()> {
+    pub fn edit_history_search(&mut self, dir: Direction) -> Result<()> {
+        let history = self.ctx.history;
         if history.is_empty() {
             return self.out.beep();
         }
-        if self.history_index == history.len() && dir == Direction::Forward
-            || self.history_index == 0 && dir == Direction::Reverse
+        if self.ctx.history_index == history.len() && dir == Direction::Forward
+            || self.ctx.history_index == 0 && dir == Direction::Reverse
         {
             return self.out.beep();
         }
         if dir == Direction::Reverse {
-            self.history_index -= 1;
+            self.ctx.history_index -= 1;
         } else {
-            self.history_index += 1;
+            self.ctx.history_index += 1;
         }
         if let Some(history_index) = history.starts_with(
             &self.line.as_str()[..self.line.pos()],
-            self.history_index,
+            self.ctx.history_index,
             dir,
         ) {
-            self.history_index = history_index;
+            self.ctx.history_index = history_index;
             let buf = history.get(history_index).unwrap();
             self.changes.borrow_mut().begin();
             self.line.update(buf, buf.len());
@@ -492,28 +535,29 @@ impl<'out, 'prompt> State<'out, 'prompt> {
     }
 
     /// Substitute the currently edited line with the first/last history entry.
-    pub fn edit_history(&mut self, history: &History, first: bool) -> Result<()> {
+    pub fn edit_history(&mut self, first: bool) -> Result<()> {
+        let history = self.ctx.history;
         if history.is_empty() {
             return Ok(());
         }
-        if self.history_index == history.len() {
+        if self.ctx.history_index == history.len() {
             if first {
                 // Save the current edited line before overwriting it
                 self.backup();
             } else {
                 return Ok(());
             }
-        } else if self.history_index == 0 && first {
+        } else if self.ctx.history_index == 0 && first {
             return Ok(());
         }
         if first {
-            self.history_index = 0;
-            let buf = history.get(self.history_index).unwrap();
+            self.ctx.history_index = 0;
+            let buf = history.get(self.ctx.history_index).unwrap();
             self.changes.borrow_mut().begin();
             self.line.update(buf, buf.len());
             self.changes.borrow_mut().end();
         } else {
-            self.history_index = history.len();
+            self.ctx.history_index = history.len();
             // Restore current edited line
             self.restore();
         }
@@ -522,11 +566,13 @@ impl<'out, 'prompt> State<'out, 'prompt> {
 }
 
 #[cfg(test)]
-pub fn init_state<'out>(
-    out: &'out mut dyn Renderer,
+pub fn init_state<'out, H: Helper>(
+    out: &'out mut <Terminal as Term>::Writer,
     line: &str,
     pos: usize,
-) -> State<'out, 'static> {
+    helper: Option<&'out H>,
+    history: &'out crate::history::History,
+) -> State<'out, 'static, H> {
     State {
         out,
         prompt: "",
@@ -534,13 +580,15 @@ pub fn init_state<'out>(
         line: LineBuffer::init(line, pos, None),
         cursor: Position::default(),
         old_rows: 0,
-        history_index: 0,
         saved_line_for_history: LineBuffer::with_capacity(100),
         byte_buffer: [0; 4],
         changes: Rc::new(RefCell::new(Changeset::new())),
-        hinter: None,
-        highlighter: None,
-        no_hint: true,
+        helper,
+        ctx: Context {
+            history,
+            history_index: 0,
+        },
+        hint: Some("hint".to_owned()),
         highlight_char: false,
     }
 }
@@ -554,38 +602,39 @@ mod test {
     #[test]
     fn edit_history_next() {
         let mut out = Sink::new();
-        let line = "current edited line";
-        let mut s = init_state(&mut out, line, 6);
         let mut history = History::new();
         history.add("line0");
         history.add("line1");
-        s.history_index = history.len();
+        let line = "current edited line";
+        let helper: Option<()> = None;
+        let mut s = init_state(&mut out, line, 6, helper.as_ref(), &history);
+        s.ctx.history_index = history.len();
 
         for _ in 0..2 {
-            s.edit_history_next(&history, false).unwrap();
+            s.edit_history_next(false).unwrap();
             assert_eq!(line, s.line.as_str());
         }
 
-        s.edit_history_next(&history, true).unwrap();
+        s.edit_history_next(true).unwrap();
         assert_eq!(line, s.saved_line_for_history.as_str());
-        assert_eq!(1, s.history_index);
+        assert_eq!(1, s.ctx.history_index);
         assert_eq!("line1", s.line.as_str());
 
         for _ in 0..2 {
-            s.edit_history_next(&history, true).unwrap();
+            s.edit_history_next(true).unwrap();
             assert_eq!(line, s.saved_line_for_history.as_str());
-            assert_eq!(0, s.history_index);
+            assert_eq!(0, s.ctx.history_index);
             assert_eq!("line0", s.line.as_str());
         }
 
-        s.edit_history_next(&history, false).unwrap();
+        s.edit_history_next(false).unwrap();
         assert_eq!(line, s.saved_line_for_history.as_str());
-        assert_eq!(1, s.history_index);
+        assert_eq!(1, s.ctx.history_index);
         assert_eq!("line1", s.line.as_str());
 
-        s.edit_history_next(&history, false).unwrap();
+        s.edit_history_next(false).unwrap();
         // assert_eq!(line, s.saved_line_for_history);
-        assert_eq!(2, s.history_index);
+        assert_eq!(2, s.ctx.history_index);
         assert_eq!(line, s.line.as_str());
     }
 }
