@@ -6,7 +6,7 @@ use super::Result;
 use crate::config::Config;
 use crate::config::EditMode;
 use crate::keys::KeyPress;
-use crate::tty::RawReader;
+use crate::tty::{RawReader, Term, Terminal};
 
 /// The number of times one command should be repeated.
 pub type RepeatCount = usize;
@@ -137,9 +137,7 @@ impl Cmd {
                     let last_insert = wrt.last_insert();
                     if let Movement::ForwardChar(0) = mvt {
                         Cmd::Replace(
-                            Movement::ForwardChar(
-                                last_insert.as_ref().map_or(0, |text| text.len()),
-                            ),
+                            Movement::ForwardChar(last_insert.as_ref().map_or(0, String::len)),
                             last_insert,
                         )
                     } else {
@@ -290,6 +288,8 @@ pub trait Refresher {
     /// Rewrite the currently edited line accordingly to the buffer content,
     /// cursor position, and number of columns of the terminal.
     fn refresh_line(&mut self) -> Result<()>;
+    /// Same as [`refresh_line`] with a specific message instead of hint
+    fn refresh_line_with_msg(&mut self, msg: Option<String>) -> Result<()>;
     /// Same as `refresh_line` but with a dynamic prompt.
     fn refresh_prompt_and_line(&mut self, prompt: &str) -> Result<()>;
     /// Vi only, switch to insert mode.
@@ -323,9 +323,9 @@ impl InputState {
     /// Parse user input into one command
     /// `single_esc_abort` is used in emacs mode on unix platform when a single
     /// esc key is expected to abort current action.
-    pub fn next_cmd<R: RawReader>(
+    pub fn next_cmd(
         &mut self,
-        rdr: &mut R,
+        rdr: &mut <Terminal as Term>::Reader,
         wrt: &mut dyn Refresher,
         single_esc_abort: bool,
     ) -> Result<Cmd> {
@@ -490,7 +490,7 @@ impl InputState {
             KeyPress::Meta('T') | KeyPress::Meta('t') => Cmd::TransposeWords(n),
             KeyPress::Meta('U') | KeyPress::Meta('u') => Cmd::UpcaseWord,
             KeyPress::Meta('Y') | KeyPress::Meta('y') => Cmd::YankPop,
-            _ => self.common(key, n, positive),
+            _ => self.common(rdr, key, n, positive)?,
         };
         debug!(target: "rustyline", "Emacs command: {:?}", cmd);
         Ok(cmd)
@@ -686,7 +686,7 @@ impl InputState {
                 Cmd::ForwardSearchHistory
             }
             KeyPress::Esc => Cmd::Noop,
-            _ => self.common(key, n, true),
+            _ => self.common(rdr, key, n, true)?,
         };
         debug!(target: "rustyline", "Vi command: {:?}", cmd);
         if cmd.is_repeatable_change() {
@@ -726,7 +726,7 @@ impl InputState {
                 wrt.done_inserting();
                 Cmd::Move(Movement::BackwardChar(1))
             }
-            _ => self.common(key, 1, true),
+            _ => self.common(rdr, key, 1, true)?,
         };
         debug!(target: "rustyline", "Vi insert: {:?}", cmd);
         if cmd.is_repeatable_change() {
@@ -827,8 +827,14 @@ impl InputState {
         })
     }
 
-    fn common(&mut self, key: KeyPress, n: RepeatCount, positive: bool) -> Cmd {
-        match key {
+    fn common<R: RawReader>(
+        &mut self,
+        rdr: &mut R,
+        key: KeyPress,
+        n: RepeatCount,
+        positive: bool,
+    ) -> Result<Cmd> {
+        Ok(match key {
             KeyPress::Home => Cmd::Move(Movement::BeginningOfLine),
             KeyPress::Left => {
                 if positive {
@@ -887,8 +893,12 @@ impl InputState {
             KeyPress::Ctrl('Z') => Cmd::Suspend,
             KeyPress::Ctrl('_') => Cmd::Undo(n),
             KeyPress::UnknownEscSeq => Cmd::Noop,
+            KeyPress::BracketedPasteStart => {
+                let paste = rdr.read_pasted_text()?;
+                Cmd::Insert(1, paste)
+            },
             _ => Cmd::Unknown,
-        }
+        })
     }
 
     fn num_args(&mut self) -> i16 {
