@@ -17,16 +17,12 @@
 //! ```
 // #![feature(non_exhaustive)]
 
-use libc;
 #[macro_use]
 extern crate log;
 
 #[cfg(unix)]
 extern crate nix;
 
-#[cfg(windows)]
-extern crate winapi;
-
 pub mod completion;
 pub mod config;
 mod edit;
@@ -70,16 +66,15 @@ use crate::line_buffer::WordAction;
 pub type Result<T> = result::Result<T, error::ReadlineError>;
 
 /// Completes the line/word
-fn complete_line<R: RawReader, C: Completer>(
-    rdr: &mut R,
-    s: &mut State<'_, '_>,
+fn complete_line<H: Helper>(
+    rdr: &mut <Terminal as Term>::Reader,
+    s: &mut State<'_, '_, H>,
     input_state: &mut InputState,
-    completer: &C,
-    highlighter: Option<&dyn Highlighter>,
     config: &Config,
 ) -> Result<Option<Cmd>> {
+    let completer = s.helper.unwrap();
     // get a list of completions
-    let (start, candidates) = completer.complete(&s.line, s.line.pos())?;
+    let (start, candidates) = completer.complete(&s.line, s.line.pos(), &s.ctx)?;
     // if no completions, we are done
     if candidates.is_empty() {
         s.out.beep()?;
@@ -178,7 +173,7 @@ fn complete_line<R: RawReader, C: Completer>(
             true
         };
         if show_completions {
-            page_completions(rdr, s, input_state, highlighter, &candidates)
+            page_completions(rdr, s, input_state, &candidates)
         } else {
             s.refresh_line()?;
             Ok(None)
@@ -189,24 +184,23 @@ fn complete_line<R: RawReader, C: Completer>(
 }
 
 /// Completes the current hint
-fn complete_hint_line(s: &mut State<'_, '_>, hinter: &dyn Hinter) -> Result<()> {
-    let hint = match hinter.hint(&s.line, s.line.pos()) {
+fn complete_hint_line<H: Helper>(s: &mut State<'_, '_, H>) -> Result<()> {
+    let hint = match s.hint.as_ref() {
         Some(hint) => hint,
         None => return Ok(()),
     };
     s.line.move_end();
-    if s.line.yank(&hint, 1).is_none() {
+    if s.line.yank(hint, 1).is_none() {
         s.out.beep()?;
     }
-    s.refresh_line()?;
+    s.refresh_line_with_msg(None)?;
     Ok(())
 }
 
-fn page_completions<R: RawReader, C: Candidate>(
-    rdr: &mut R,
-    s: &mut State<'_, '_>,
+fn page_completions<C: Candidate, H: Helper>(
+    rdr: &mut <Terminal as Term>::Reader,
+    s: &mut State<'_, '_, H>,
     input_state: &mut InputState,
-    highlighter: Option<&dyn Highlighter>,
     candidates: &[C],
 ) -> Result<Option<Cmd>> {
     use std::cmp;
@@ -262,7 +256,7 @@ fn page_completions<R: RawReader, C: Candidate>(
             if i < candidates.len() {
                 let candidate = &candidates[i].display();
                 let width = candidate.width();
-                if let Some(highlighter) = highlighter {
+                if let Some(highlighter) = s.highlighter() {
                     ab.push_str(&highlighter.highlight_candidate(candidate, CompletionType::List));
                 } else {
                     ab.push_str(candidate);
@@ -282,9 +276,9 @@ fn page_completions<R: RawReader, C: Candidate>(
 }
 
 /// Incremental search
-fn reverse_incremental_search<R: RawReader>(
-    rdr: &mut R,
-    s: &mut State<'_, '_>,
+fn reverse_incremental_search<H: Helper>(
+    rdr: &mut <Terminal as Term>::Reader,
+    s: &mut State<'_, '_, H>,
     input_state: &mut InputState,
     history: &History,
 ) -> Result<Option<Cmd>> {
@@ -376,24 +370,16 @@ fn readline_edit<H: Helper>(
     editor: &mut Editor<H>,
     original_mode: &tty::Mode,
 ) -> Result<String> {
-    let completer = editor.helper.as_ref();
-    let hinter = editor.helper.as_ref().map(|h| h as &dyn Hinter);
-    let highlighter = if editor.term.colors_enabled() {
-        editor.helper.as_ref().map(|h| h as &dyn Highlighter)
-    } else {
-        None
-    };
+    let helper = editor.helper.as_ref();
 
     let mut stdout = editor.term.create_writer();
 
     editor.reset_kill_ring(); // TODO recreate a new kill ring vs Arc<Mutex<KillRing>>
-    let mut s = State::new(
-        &mut stdout,
-        prompt,
-        editor.history.len(),
-        hinter,
-        highlighter,
-    );
+    let ctx = Context {
+        history: &editor.history,
+        history_index: editor.history.len(),
+    };
+    let mut s = State::new(&mut stdout, prompt, helper, ctx);
     let mut input_state = InputState::new(&editor.config, Arc::clone(&editor.custom_bindings));
 
     s.line.set_delete_listener(editor.kill_ring.clone());
@@ -404,9 +390,9 @@ fn readline_edit<H: Helper>(
             .update((left.to_owned() + right).as_ref(), left.len());
     }
 
-    s.refresh_line()?;
-
     let mut rdr = editor.term.create_reader(&editor.config)?;
+    s.move_cursor_at_leftmost(&mut rdr)?;
+    s.refresh_line()?;
 
     loop {
         let rc = s.next_cmd(&mut input_state, &mut rdr, false);
@@ -417,26 +403,17 @@ fn readline_edit<H: Helper>(
         }
 
         // autocomplete
-        if cmd == Cmd::Complete && completer.is_some() {
-            let next = complete_line(
-                &mut rdr,
-                &mut s,
-                &mut input_state,
-                completer.unwrap(),
-                highlighter,
-                &editor.config,
-            )?;
-            if next.is_some() {
-                cmd = next.unwrap();
+        if cmd == Cmd::Complete && s.helper.is_some() {
+            let next = complete_line(&mut rdr, &mut s, &mut input_state, &editor.config)?;
+            if let Some(next) = next {
+                cmd = next;
             } else {
                 continue;
             }
         }
 
         if let Cmd::CompleteHint = cmd {
-            if hinter.is_some() {
-                complete_hint_line(&mut s, hinter.unwrap())?;
-            }
+            complete_hint_line(&mut s)?;
             continue;
         }
 
@@ -452,8 +429,8 @@ fn readline_edit<H: Helper>(
             // Search history backward
             let next =
                 reverse_incremental_search(&mut rdr, &mut s, &mut input_state, &editor.history)?;
-            if next.is_some() {
-                cmd = next.unwrap();
+            if let Some(next) = next {
+                cmd = next;
             } else {
                 continue;
             }
@@ -507,18 +484,14 @@ fn readline_edit<H: Helper>(
             }
             Cmd::NextHistory => {
                 // Fetch the next command from the history list.
-                s.edit_history_next(&editor.history, false)?
+                s.edit_history_next(false)?
             }
             Cmd::PreviousHistory => {
                 // Fetch the previous command from the history list.
-                s.edit_history_next(&editor.history, true)?
-            }
-            Cmd::HistorySearchBackward => {
-                s.edit_history_search(&editor.history, Direction::Reverse)?
-            }
-            Cmd::HistorySearchForward => {
-                s.edit_history_search(&editor.history, Direction::Forward)?
+                s.edit_history_next(true)?
             }
+            Cmd::HistorySearchBackward => s.edit_history_search(Direction::Reverse)?,
+            Cmd::HistorySearchForward => s.edit_history_search(Direction::Forward)?,
             Cmd::TransposeChars => {
                 // Exchange the char before cursor with the character at cursor.
                 s.edit_transpose_chars()?
@@ -527,7 +500,7 @@ fn readline_edit<H: Helper>(
             Cmd::QuotedInsert => {
                 // Quoted insert
                 let c = rdr.next_char()?;
-                s.edit_insert(c, 1)? // FIXME
+                s.edit_insert(c, 1)?
             }
             Cmd::Yank(n, anchor) => {
                 // retrieve (yank) last item killed
@@ -542,7 +515,6 @@ fn readline_edit<H: Helper>(
                     kill_ring.kill(&text, Mode::Append)
                 }
             }
-            // TODO CTRL-_ // undo
             Cmd::AcceptLine => {
                 #[cfg(test)]
                 {
@@ -550,21 +522,20 @@ fn readline_edit<H: Helper>(
                 }
                 // Accept the line regardless of where the cursor is.
                 s.edit_move_end()?;
-                if s.hinter.is_some() {
+                if s.has_hint() {
                     // Force a refresh without hints to leave the previous
                     // line as the user typed it after a newline.
-                    s.hinter = None;
-                    s.refresh_line()?;
+                    s.refresh_line_with_msg(None)?;
                 }
                 break;
             }
             Cmd::BeginningOfHistory => {
                 // move to first entry in history
-                s.edit_history(&editor.history, true)?
+                s.edit_history(true)?
             }
             Cmd::EndOfHistory => {
                 // move to last entry in history
-                s.edit_history(&editor.history, false)?
+                s.edit_history(false)?
             }
             Cmd::Move(Movement::BackwardWord(n, word_def)) => {
                 // move backwards one word
@@ -602,11 +573,9 @@ fn readline_edit<H: Helper>(
             }
             Cmd::Move(Movement::ViCharSearch(n, cs)) => s.edit_move_to(cs, n)?,
             Cmd::Undo(n) => {
-                s.line.remove_change_listener();
                 if s.changes.borrow_mut().undo(&mut s.line, n) {
                     s.refresh_line()?;
                 }
-                s.line.set_change_listener(s.changes.clone());
             }
             Cmd::Interrupt => {
                 return Err(error::ReadlineError::Interrupted);
@@ -652,7 +621,7 @@ fn readline_raw<H: Helper>(
     let user_input = readline_edit(prompt, initial, editor, &original_mode);
     if editor.config.auto_add_history() {
         if let Ok(ref line) = user_input {
-            editor.add_history_entry(line.as_ref());
+            editor.add_history_entry(line.as_str());
         }
     }
     drop(guard); // disable_raw_mode(original_mode)?;
@@ -686,6 +655,26 @@ where
 
 impl Helper for () {}
 
+impl<'h, H: ?Sized + Helper> Helper for &'h H {}
+
+/// Completion/suggestion context
+pub struct Context<'h> {
+    history: &'h History,
+    history_index: usize,
+}
+
+impl<'h> Context<'h> {
+    /// Return an immutable reference to the history object.
+    pub fn history(&self) -> &History {
+        &self.history
+    }
+
+    /// The history index we are currently editing
+    pub fn history_index(&self) -> usize {
+        self.history_index
+    }
+}
+
 /// Line editor
 pub struct Editor<H: Helper> {
     term: Terminal,
@@ -705,7 +694,11 @@ impl<H: Helper> Editor<H> {
 
     /// Create an editor with a specific configuration.
     pub fn with_config(config: Config) -> Self {
-        let term = Terminal::new(config.color_mode(), config.output_stream());
+        let term = Terminal::new(
+            config.color_mode(),
+            config.output_stream(),
+            config.tab_stop(),
+        );
         Self {
             term,
             history: History::with_config(config),
@@ -803,14 +796,20 @@ impl<H: Helper> Editor<H> {
 
     /// Bind a sequence to a command.
     pub fn bind_sequence(&mut self, key_seq: KeyPress, cmd: Cmd) -> Option<Cmd> {
-        let mut bindings = self.custom_bindings.write().unwrap();
-        bindings.insert(key_seq, cmd)
+        if let Ok(mut bindings) = self.custom_bindings.write() {
+            bindings.insert(key_seq, cmd)
+        } else {
+            None
+        }
     }
 
     /// Remove a binding for the given sequence.
     pub fn unbind_sequence(&mut self, key_seq: KeyPress) -> Option<Cmd> {
-        let mut bindings = self.custom_bindings.write().unwrap();
-        bindings.remove(&key_seq)
+        if let Ok(mut bindings) = self.custom_bindings.write() {
+            bindings.remove(&key_seq)
+        } else {
+            None
+        }
     }
 
     /// ```
@@ -838,6 +837,17 @@ impl<H: Helper> Editor<H> {
         let mut kill_ring = self.kill_ring.lock().unwrap();
         kill_ring.reset();
     }
+
+    /// If output stream is a tty, this function returns its width and height as
+    /// a number of characters.
+    pub fn dimensions(&mut self) -> Option<(usize, usize)> {
+        if self.term.is_output_tty() {
+            let out = self.term.create_writer();
+            Some((out.get_columns(), out.get_rows()))
+        } else {
+            None
+        }
+    }
 }
 
 impl<H: Helper> config::Configurer for Editor<H> {
