@@ -31,6 +31,10 @@ impl<'a> RawReader for Iter<'a, KeyPress> {
     fn next_char(&mut self) -> Result<char> {
         unimplemented!();
     }
+
+    fn read_pasted_text(&mut self) -> Result<String> {
+        unimplemented!()
+    }
 }
 
 impl RawReader for IntoIter<KeyPress> {
@@ -49,6 +53,10 @@ impl RawReader for IntoIter<KeyPress> {
             _ => unimplemented!(),
         }
     }
+
+    fn read_pasted_text(&mut self) -> Result<String> {
+        unimplemented!()
+    }
 }
 
 pub struct Sink {}
@@ -60,6 +68,8 @@ impl Sink {
 }
 
 impl Renderer for Sink {
+    type Reader = IntoIter<KeyPress>;
+
     fn move_cursor(&mut self, _: Position, _: Position) -> Result<()> {
         Ok(())
     }
@@ -68,8 +78,9 @@ impl Renderer for Sink {
         &mut self,
         _: &str,
         prompt_size: Position,
+        _: bool,
         line: &LineBuffer,
-        hint: Option<String>,
+        hint: Option<&str>,
         _: usize,
         _: usize,
         _: Option<&dyn Highlighter>,
@@ -113,6 +124,14 @@ impl Renderer for Sink {
     fn get_rows(&self) -> usize {
         24
     }
+
+    fn colors_enabled(&self) -> bool {
+        false
+    }
+
+    fn move_cursor_at_leftmost(&mut self, _: &mut IntoIter<KeyPress>) -> Result<()> {
+        Ok(())
+    }
 }
 
 pub type Terminal = DummyTerminal;
@@ -129,7 +148,7 @@ impl Term for DummyTerminal {
     type Reader = IntoIter<KeyPress>;
     type Writer = Sink;
 
-    fn new(color_mode: ColorMode, _stream: OutputStreamType) -> DummyTerminal {
+    fn new(color_mode: ColorMode, _stream: OutputStreamType, _tab_stop: usize) -> DummyTerminal {
         DummyTerminal {
             keys: Vec::new(),
             cursor: 0,
@@ -147,7 +166,7 @@ impl Term for DummyTerminal {
         true
     }
 
-    fn colors_enabled(&self) -> bool {
+    fn is_output_tty(&self) -> bool {
         false
     }
 
