@@ -7,7 +7,7 @@ use std::sync::atomic;
 
 use libc;
 use nix;
-use nix::poll::{self, EventFlags};
+use nix::poll::{self, PollFlags};
 use nix::sys::signal;
 use nix::sys::termios;
 use nix::sys::termios::SetArg;
@@ -27,6 +27,9 @@ const STDIN_FILENO: libc::c_int = libc::STDIN_FILENO;
 /// Unsupported Terminals that don't support RAW mode
 static UNSUPPORTED_TERM: [&'static str; 3] = ["dumb", "cons25", "emacs"];
 
+const BRACKETED_PASTE_ON: &[u8] = b"\x1b[?2004h";
+const BRACKETED_PASTE_OFF: &[u8] = b"\x1b[?2004l";
+
 impl AsRawFd for OutputStreamType {
     fn as_raw_fd(&self) -> RawFd {
         match self {
@@ -72,13 +75,22 @@ fn is_a_tty(fd: libc::c_int) -> bool {
     unsafe { libc::isatty(fd) != 0 }
 }
 
-#[cfg(not(test))]
-pub type Mode = termios::Termios;
+pub struct PosixMode {
+    termios: termios::Termios,
+    out: Option<OutputStreamType>,
+}
 
-impl RawMode for termios::Termios {
+#[cfg(not(test))]
+pub type Mode = PosixMode;
+
+impl RawMode for PosixMode {
     /// Disable RAW mode for the terminal.
     fn disable_raw_mode(&self) -> Result<()> {
-        termios::tcsetattr(STDIN_FILENO, SetArg::TCSADRAIN, self)?;
+        termios::tcsetattr(STDIN_FILENO, SetArg::TCSADRAIN, &self.termios)?;
+        // disable bracketed paste
+        if let Some(out) = self.out {
+            write_and_flush(out, BRACKETED_PASTE_OFF)?;
+        }
         Ok(())
     }
 }
@@ -207,6 +219,7 @@ impl PosixRawReader {
     }
 
     /// Handle ESC [ <seq2:digit> escape sequences
+    #[allow(clippy::cognitive_complexity)]
     fn extended_escape(&mut self, seq2: char) -> Result<KeyPress> {
         let seq3 = self.next_char()?;
         if seq3 == '~' {
@@ -256,6 +269,23 @@ impl PosixRawReader {
                            "unsupported esc sequence: ESC [ {}{} ; {:?}", seq2, seq3, seq5);
                 }
                 Ok(KeyPress::UnknownEscSeq)
+            } else if seq4.is_digit(10) {
+                let seq5 = self.next_char()?;
+                if seq5 == '~' {
+                    Ok(match (seq2, seq3, seq4) {
+                        ('2', '0', '0') => KeyPress::BracketedPasteStart,
+                        ('2', '0', '1') => KeyPress::BracketedPasteEnd,
+                        _ => {
+                            debug!(target: "rustyline",
+                                   "unsupported esc sequence: ESC [ {}{}{}~", seq2, seq3, seq4);
+                            KeyPress::UnknownEscSeq
+                        }
+                    })
+                } else {
+                    debug!(target: "rustyline",
+                           "unsupported esc sequence: ESC [ {}{}{} {}", seq2, seq3, seq4, seq5);
+                    Ok(KeyPress::UnknownEscSeq)
+                }
             } else {
                 debug!(target: "rustyline",
                        "unsupported esc sequence: ESC [ {}{} {:?}", seq2, seq3, seq4);
@@ -343,7 +373,7 @@ impl RawReader for PosixRawReader {
             } else {
                 self.timeout_ms
             };
-            let mut fds = [poll::PollFd::new(STDIN_FILENO, EventFlags::POLLIN)];
+            let mut fds = [poll::PollFd::new(STDIN_FILENO, PollFlags::POLLIN)];
             match poll::poll(&mut fds, timeout_ms) {
                 Ok(n) if n == 0 => {
                     // single escape
@@ -370,11 +400,31 @@ impl RawReader for PosixRawReader {
             self.parser.advance(&mut self.receiver, b);
             if !self.receiver.valid {
                 return Err(error::ReadlineError::Utf8Error);
-            } else if self.receiver.c.is_some() {
-                return Ok(self.receiver.c.take().unwrap());
+            } else if let Some(c) = self.receiver.c.take() {
+                return Ok(c);
             }
         }
     }
+
+    fn read_pasted_text(&mut self) -> Result<String> {
+        let mut buffer = String::new();
+        loop {
+            match self.next_char()? {
+                '\x1b' => {
+                    let key = self.escape_sequence()?;
+                    if key == KeyPress::BracketedPasteEnd {
+                        break;
+                    } else {
+                        continue; // TODO validate
+                    }
+                }
+                c => buffer.push(c),
+            };
+        }
+        let buffer = buffer.replace("\r\n", "\n");
+        let buffer = buffer.replace("\r", "\n");
+        Ok(buffer)
+    }
 }
 
 impl Receiver for Utf8 {
@@ -396,66 +446,73 @@ pub struct PosixRenderer {
     out: OutputStreamType,
     cols: usize, // Number of columns in terminal
     buffer: String,
+    tab_stop: usize,
+    colors_enabled: bool,
 }
 
 impl PosixRenderer {
-    fn new(out: OutputStreamType) -> Self {
+    fn new(out: OutputStreamType, tab_stop: usize, colors_enabled: bool) -> Self {
         let (cols, _) = get_win_size(&out);
         Self {
             out,
             cols,
             buffer: String::with_capacity(1024),
+            tab_stop,
+            colors_enabled,
         }
     }
 }
 
 impl Renderer for PosixRenderer {
+    type Reader = PosixRawReader;
+
     fn move_cursor(&mut self, old: Position, new: Position) -> Result<()> {
         use std::fmt::Write;
-        let mut ab = String::new();
+        self.buffer.clear();
         if new.row > old.row {
             // move down
             let row_shift = new.row - old.row;
             if row_shift == 1 {
-                ab.push_str("\x1b[B");
+                self.buffer.push_str("\x1b[B");
             } else {
-                write!(ab, "\x1b[{}B", row_shift).unwrap();
+                write!(self.buffer, "\x1b[{}B", row_shift).unwrap();
             }
         } else if new.row < old.row {
             // move up
             let row_shift = old.row - new.row;
             if row_shift == 1 {
-                ab.push_str("\x1b[A");
+                self.buffer.push_str("\x1b[A");
             } else {
-                write!(ab, "\x1b[{}A", row_shift).unwrap();
+                write!(self.buffer, "\x1b[{}A", row_shift).unwrap();
             }
         }
         if new.col > old.col {
             // move right
             let col_shift = new.col - old.col;
             if col_shift == 1 {
-                ab.push_str("\x1b[C");
+                self.buffer.push_str("\x1b[C");
             } else {
-                write!(ab, "\x1b[{}C", col_shift).unwrap();
+                write!(self.buffer, "\x1b[{}C", col_shift).unwrap();
             }
         } else if new.col < old.col {
             // move left
             let col_shift = old.col - new.col;
             if col_shift == 1 {
-                ab.push_str("\x1b[D");
+                self.buffer.push_str("\x1b[D");
             } else {
-                write!(ab, "\x1b[{}D", col_shift).unwrap();
+                write!(self.buffer, "\x1b[{}D", col_shift).unwrap();
             }
         }
-        self.write_and_flush(ab.as_bytes())
+        self.write_and_flush(self.buffer.as_bytes())
     }
 
     fn refresh_line(
         &mut self,
         prompt: &str,
         prompt_size: Position,
+        default_prompt: bool,
         line: &LineBuffer,
-        hint: Option<String>,
+        hint: Option<&str>,
         current_row: usize,
         old_rows: usize,
         highlighter: Option<&dyn Highlighter>,
@@ -484,7 +541,8 @@ impl Renderer for PosixRenderer {
 
         if let Some(highlighter) = highlighter {
             // display the prompt
-            self.buffer.push_str(&highlighter.highlight_prompt(prompt));
+            self.buffer
+                .push_str(&highlighter.highlight_prompt(prompt, default_prompt));
             // display the input line
             self.buffer
                 .push_str(&highlighter.highlight(line, line.pos()));
@@ -525,17 +583,7 @@ impl Renderer for PosixRenderer {
     }
 
     fn write_and_flush(&self, buf: &[u8]) -> Result<()> {
-        match self.out {
-            OutputStreamType::Stdout => {
-                io::stdout().write_all(buf)?;
-                io::stdout().flush()?;
-            }
-            OutputStreamType::Stderr => {
-                io::stderr().write_all(buf)?;
-                io::stderr().flush()?;
-            }
-        }
-        Ok(())
+        write_and_flush(self.out, buf)
     }
 
     /// Control characters are treated as having zero width.
@@ -549,7 +597,11 @@ impl Renderer for PosixRenderer {
                 pos.col = 0;
                 continue;
             }
-            let cw = width(c, &mut esc_seq);
+            let cw = if c == "\t" {
+                self.tab_stop - (pos.col % self.tab_stop)
+            } else {
+                width(c, &mut esc_seq)
+            };
             pos.col += cw;
             if pos.col > self.cols {
                 pos.row += 1;
@@ -589,6 +641,46 @@ impl Renderer for PosixRenderer {
         let (_, rows) = get_win_size(&self.out);
         rows
     }
+
+    fn colors_enabled(&self) -> bool {
+        self.colors_enabled
+    }
+
+    fn move_cursor_at_leftmost(&mut self, rdr: &mut PosixRawReader) -> Result<()> {
+        /* Report cursor location */
+        self.write_and_flush(b"\x1b[6n")?;
+        /* Read the response: ESC [ rows ; cols R */
+        if rdr.next_char()? != '\x1b' {
+            return Err(error::ReadlineError::from(io::ErrorKind::InvalidData));
+        }
+        if rdr.next_char()? != '[' {
+            return Err(error::ReadlineError::from(io::ErrorKind::InvalidData));
+        }
+        read_digits_until(rdr, ';')?;
+        let col = read_digits_until(rdr, 'R')?;
+        debug!(target: "rustyline", "initial cursor location: {:?}", col);
+        if col != 1 {
+            self.write_and_flush(b"\n")?;
+        }
+        Ok(())
+    }
+}
+
+fn read_digits_until(rdr: &mut PosixRawReader, sep: char) -> Result<u32> {
+    let mut num: u32 = 0;
+    loop {
+        match rdr.next_char()? {
+            digit @ '0'..='9' => {
+                num = num
+                    .saturating_mul(10)
+                    .saturating_add(digit.to_digit(10).unwrap());
+                continue;
+            }
+            c if c == sep => break,
+            _ => return Err(error::ReadlineError::from(io::ErrorKind::InvalidData)),
+        }
+    }
+    Ok(num)
 }
 
 static SIGWINCH_ONCE: sync::Once = sync::ONCE_INIT;
@@ -620,20 +712,32 @@ pub struct PosixTerminal {
     stdstream_isatty: bool,
     pub(crate) color_mode: ColorMode,
     stream_type: OutputStreamType,
+    tab_stop: usize,
+}
+
+impl PosixTerminal {
+    fn colors_enabled(&self) -> bool {
+        match self.color_mode {
+            ColorMode::Enabled => self.stdstream_isatty,
+            ColorMode::Forced => true,
+            ColorMode::Disabled => false,
+        }
+    }
 }
 
 impl Term for PosixTerminal {
-    type Mode = termios::Termios;
+    type Mode = PosixMode;
     type Reader = PosixRawReader;
     type Writer = PosixRenderer;
 
-    fn new(color_mode: ColorMode, stream_type: OutputStreamType) -> Self {
+    fn new(color_mode: ColorMode, stream_type: OutputStreamType, tab_stop: usize) -> Self {
         let term = Self {
             unsupported: is_unsupported_term(),
             stdin_isatty: is_a_tty(STDIN_FILENO),
             stdstream_isatty: is_a_tty(stream_type.as_raw_fd()),
             color_mode,
             stream_type,
+            tab_stop,
         };
         if !term.unsupported && term.stdin_isatty && term.stdstream_isatty {
             install_sigwinch_handler();
@@ -654,13 +758,8 @@ impl Term for PosixTerminal {
         self.stdin_isatty
     }
 
-    /// Check if output supports colors.
-    fn colors_enabled(&self) -> bool {
-        match self.color_mode {
-            ColorMode::Enabled => self.stdstream_isatty,
-            ColorMode::Forced => true,
-            ColorMode::Disabled => false,
-        }
+    fn is_output_tty(&self) -> bool {
+        self.stdstream_isatty
     }
 
     // Interactive loop:
@@ -692,7 +791,18 @@ impl Term for PosixTerminal {
         raw.control_chars[SpecialCharacterIndices::VMIN as usize] = 1; // One character-at-a-time input
         raw.control_chars[SpecialCharacterIndices::VTIME as usize] = 0; // with blocking read
         termios::tcsetattr(STDIN_FILENO, SetArg::TCSADRAIN, &raw)?;
-        Ok(original_mode)
+
+        // enable bracketed paste
+        let out = if let Err(e) = write_and_flush(self.stream_type, BRACKETED_PASTE_ON) {
+            debug!(target: "rustyline", "Cannot enable bracketed paste: {}", e);
+            None
+        } else {
+            Some(self.stream_type)
+        };
+        Ok(PosixMode {
+            termios: original_mode,
+            out,
+        })
     }
 
     /// Create a RAW reader
@@ -701,7 +811,7 @@ impl Term for PosixTerminal {
     }
 
     fn create_writer(&self) -> PosixRenderer {
-        PosixRenderer::new(self.stream_type)
+        PosixRenderer::new(self.stream_type, self.tab_stop, self.colors_enabled())
     }
 }
 
@@ -713,6 +823,20 @@ pub fn suspend() -> Result<()> {
     Ok(())
 }
 
+fn write_and_flush(out: OutputStreamType, buf: &[u8]) -> Result<()> {
+    match out {
+        OutputStreamType::Stdout => {
+            io::stdout().write_all(buf)?;
+            io::stdout().flush()?;
+        }
+        OutputStreamType::Stderr => {
+            io::stderr().write_all(buf)?;
+            io::stderr().flush()?;
+        }
+    }
+    Ok(())
+}
+
 #[cfg(test)]
 mod test {
     use super::{Position, PosixRenderer, PosixTerminal, Renderer};
@@ -721,7 +845,7 @@ mod test {
     #[test]
     #[ignore]
     fn prompt_with_ansi_escape_codes() {
-        let out = PosixRenderer::new(OutputStreamType::Stdout);
+        let out = PosixRenderer::new(OutputStreamType::Stdout, 4, true);
         let pos = out.calculate_position("\x1b[1;32m>>\x1b[0m ", Position::default());
         assert_eq!(3, pos.col);
         assert_eq!(0, pos.row);
