@@ -1,7 +1,6 @@
 //! Windows specific definitions
 use std::io::{self, Write};
 use std::mem;
-use std::os::windows::io::{AsRawHandle, RawHandle};
 use std::sync::atomic;
 
 use unicode_width::UnicodeWidthChar;
@@ -97,7 +96,7 @@ pub struct ConsoleRawReader {
 }
 
 impl ConsoleRawReader {
-    pub fn new() -> Result<ConsoleRawReader> {
+    pub fn create() -> Result<ConsoleRawReader> {
         let handle = get_std_handle(STDIN_FILENO)?;
         Ok(ConsoleRawReader { handle })
     }
@@ -147,7 +146,7 @@ impl RawReader for ConsoleRawReader {
 
             let utf16 = unsafe { *key_event.uChar.UnicodeChar() };
             if utf16 == 0 {
-                match key_event.wVirtualKeyCode as i32 {
+                match i32::from(key_event.wVirtualKeyCode) {
                     winuser::VK_LEFT => {
                         return Ok(if ctrl {
                             KeyPress::ControlLeft
@@ -218,10 +217,12 @@ impl RawReader for ConsoleRawReader {
                 } else {
                     decode_utf16([surrogate, utf16].iter().cloned()).next()
                 };
-                if orc.is_none() {
+                let rc = if let Some(rc) = orc {
+                    rc
+                } else {
                     return Err(error::ReadlineError::Eof);
-                }
-                let c = orc.unwrap()?;
+                };
+                let c = rc?;
                 if meta {
                     return Ok(KeyPress::Meta(c));
                 } else {
@@ -236,6 +237,10 @@ impl RawReader for ConsoleRawReader {
             }
         }
     }
+
+    fn read_pasted_text(&mut self) -> Result<String> {
+        unimplemented!()
+    }
 }
 
 pub struct ConsoleRenderer {
@@ -243,10 +248,11 @@ pub struct ConsoleRenderer {
     handle: HANDLE,
     cols: usize, // Number of columns in terminal
     buffer: String,
+    colors_enabled: bool,
 }
 
 impl ConsoleRenderer {
-    fn new(handle: HANDLE, out: OutputStreamType) -> ConsoleRenderer {
+    fn new(handle: HANDLE, out: OutputStreamType, colors_enabled: bool) -> ConsoleRenderer {
         // Multi line editing is enabled by ENABLE_WRAP_AT_EOL_OUTPUT mode
         let (cols, _) = get_win_size(handle);
         ConsoleRenderer {
@@ -254,6 +260,7 @@ impl ConsoleRenderer {
             handle,
             cols,
             buffer: String::with_capacity(1024),
+            colors_enabled,
         }
     }
 
@@ -282,6 +289,8 @@ impl ConsoleRenderer {
 }
 
 impl Renderer for ConsoleRenderer {
+    type Reader = ConsoleRawReader;
+
     fn move_cursor(&mut self, old: Position, new: Position) -> Result<()> {
         let mut info = self.get_console_screen_buffer_info()?;
         if new.row > old.row {
@@ -301,11 +310,12 @@ impl Renderer for ConsoleRenderer {
         &mut self,
         prompt: &str,
         prompt_size: Position,
+        default_prompt: bool,
         line: &LineBuffer,
-        hint: Option<String>,
+        hint: Option<&str>,
         current_row: usize,
         old_rows: usize,
-        highlighter: Option<&Highlighter>,
+        highlighter: Option<&dyn Highlighter>,
     ) -> Result<(Position, Position)> {
         // calculate the position of the end of the input line
         let end_pos = self.calculate_position(line, prompt_size);
@@ -325,7 +335,8 @@ impl Renderer for ConsoleRenderer {
         if let Some(highlighter) = highlighter {
             // TODO handle ansi escape code (SetConsoleTextAttribute)
             // display the prompt
-            self.buffer.push_str(&highlighter.highlight_prompt(prompt));
+            self.buffer
+                .push_str(&highlighter.highlight_prompt(prompt, default_prompt));
             // display the input line
             self.buffer
                 .push_str(&highlighter.highlight(line, line.pos()));
@@ -424,9 +435,25 @@ impl Renderer for ConsoleRenderer {
         let (_, rows) = get_win_size(self.handle);
         rows
     }
+
+    fn colors_enabled(&self) -> bool {
+        self.colors_enabled
+    }
+
+    fn move_cursor_at_leftmost(&mut self, _: &mut ConsoleRawReader) -> Result<()> {
+        self.write_and_flush(b"")?; // we must do this otherwise the cursor position is not reported correctly
+        let mut info = self.get_console_screen_buffer_info()?;
+        if info.dwCursorPosition.X == 0 {
+            return Ok(());
+        }
+        debug!(target: "rustyline", "initial cursor location: {:?}, {:?}", info.dwCursorPosition.X, info.dwCursorPosition.Y);
+        info.dwCursorPosition.X = 0;
+        info.dwCursorPosition.Y += 1;
+        self.set_console_cursor_position(info.dwCursorPosition)
+    }
 }
 
-static SIGWINCH: atomic::AtomicBool = atomic::ATOMIC_BOOL_INIT;
+static SIGWINCH: atomic::AtomicBool = atomic::AtomicBool::new(false);
 
 #[cfg(not(test))]
 pub type Terminal = Console;
@@ -442,14 +469,23 @@ pub struct Console {
     stream_type: OutputStreamType,
 }
 
-impl Console {}
+impl Console {
+    fn colors_enabled(&self) -> bool {
+        // TODO ANSI Colors & Windows <10
+        match self.color_mode {
+            ColorMode::Enabled => self.stdstream_isatty && self.ansi_colors_supported,
+            ColorMode::Forced => true,
+            ColorMode::Disabled => false,
+        }
+    }
+}
 
 impl Term for Console {
     type Mode = ConsoleMode;
     type Reader = ConsoleRawReader;
     type Writer = ConsoleRenderer;
 
-    fn new(color_mode: ColorMode, stream_type: OutputStreamType) -> Console {
+    fn new(color_mode: ColorMode, stream_type: OutputStreamType, _tab_stop: usize) -> Console {
         use std::ptr;
         let stdin_handle = get_std_handle(STDIN_FILENO);
         let stdin_isatty = match stdin_handle {
@@ -493,13 +529,8 @@ impl Term for Console {
         self.stdin_isatty
     }
 
-    fn colors_enabled(&self) -> bool {
-        // TODO ANSI Colors & Windows <10
-        match self.color_mode {
-            ColorMode::Enabled => self.stdstream_isatty && self.ansi_colors_supported,
-            ColorMode::Forced => true,
-            ColorMode::Disabled => false,
-        }
+    fn is_output_tty(&self) -> bool {
+        self.stdstream_isatty
     }
 
     // pub fn install_sigwinch_handler(&mut self) {
@@ -535,6 +566,10 @@ impl Term for Console {
                 let raw = original_stdstream_mode | wincon::ENABLE_VIRTUAL_TERMINAL_PROCESSING;
                 self.ansi_colors_supported =
                     unsafe { consoleapi::SetConsoleMode(self.stdstream_handle, raw) != 0 };
+                debug!(target: "rustyline", "ansi_colors_supported: {}", self.ansi_colors_supported);
+            } else {
+                debug!(target: "rustyline", "ANSI colors already enabled");
+                self.ansi_colors_supported = true;
             }
             Some(original_stdstream_mode)
         } else {
@@ -550,11 +585,15 @@ impl Term for Console {
     }
 
     fn create_reader(&self, _: &Config) -> Result<ConsoleRawReader> {
-        ConsoleRawReader::new()
+        ConsoleRawReader::create()
     }
 
     fn create_writer(&self) -> ConsoleRenderer {
-        ConsoleRenderer::new(self.stdstream_handle, self.stream_type)
+        ConsoleRenderer::new(
+            self.stdstream_handle,
+            self.stream_type,
+            self.colors_enabled(),
+        )
     }
 }
 
